"""Script with functions that converts one variable to another.

The converting functions are made through trial and error by matching the synthetically
created variables with one real data set used in CESM2.
"""

from typing import Tuple

import numpy as np


def vei_to_totalemission(veis: np.ndarray) -> np.ndarray:
    """Convert from VEI to Total_Emission.

    Parameters
    ----------
    veis: np.ndarray
        The array of VEI that should be converted to Total_Emission

    Returns
    -------
    np.ndarray
        The resulting Total_Emission's array as a 1D numpy array

    Raises
    ------
    ValueError
        If the input is not an array of float32.
    """
    if veis.dtype != np.int8:
        raise ValueError(f"{veis.dtype} = . Need int8.")
    return 1e-2 * 3 ** (
        np.random.normal(0.1, 1.0, size=len(veis)).astype(np.float32) + veis
    )


def totalemission_to_vei(tes: np.ndarray) -> np.ndarray:
    """Convert from Total_Emission to VEI.

    All values in the VEI array are mapped to [0, 7].

    Parameters
    ----------
    tes: np.ndarray
        The array of Total_Emission that should be converted to VEI

    Returns
    -------
    np.ndarray
        The resulting VEI's array as a 1D numpy array

    Raises
    ------
    ZeroDivisionError
        If `tes` has non-positive values.
    """
    if any(tes <= 0):
        raise ZeroDivisionError(
            "Don't use total emission with zeros. We're doing a log."
        )
    base = 4
    veis = np.log(tes) / np.log(base)
    veis -= np.min(veis) + 0.2
    veis = veis.astype(np.int8) % 7

    return veis


def vei_to_injectionheights(veis: np.ndarray) -> Tuple[np.ndarray, np.ndarray]:
    """Convert VEI to minimum and maximum injection heights.

    When we set the minimum and maximum injection heights we should make sure the minimum
    value really is less than the maximum value, hence the for loop at the end.  The
    variables 'scale_min' and 'scale_max' were adjusted based on the helper script
    'inspect_X-vs-Y.py'.

    Parameters
    ----------
    veis: np.ndarray
        The array of VEI that should be converted to Total_Emission

    Returns
    -------
    tuple[np.ndarray, np.ndarray]
        The resulting minimum and maximum injection heights as a 1D numpy arrays.
        The first element is the lower bound, second is upper bound.
    """
    scale_max = np.array([80 if v > 3 else 10 for v in veis])
    scale_min = np.array([20 if v > 3 else 1 for v in veis])
    mxihs = np.abs(
        np.random.normal(1, 2.0, size=len(veis)).astype(np.float32) * scale_max
    )
    miihs = np.abs(
        np.random.normal(1, 2.0, size=len(veis)).astype(np.float32) * scale_min
    )
    # Finally we make sure the lower bound is less than 30:
    idx_min = miihs > 30
    miihs[idx_min] = 18
    idx_max = mxihs > 30
    mxihs[idx_max] = 20

    for idx, (i, j) in enumerate(zip(miihs, mxihs)):
        miihs[idx] = min(i, j)
        mxihs[idx] = max(i, j)
    mxihs = mxihs.astype(np.float32)
    miihs = miihs.astype(np.float32)
    return miihs, mxihs



"""
Standards for figures and plots.

Use the set_rcparams_dynamo routine to get a sane default configuration
for production quality plots.

The defaults should be used for most purposes.


Some guidelines when creating plots:

The default colour palette of matplotlib is good for red-green colour-blindness,
but may be bad for other types.

For scans where there the variable can be sorted by size,
use the following plot styles:

Variable    Symbol    Color
Large        ^         C0 #1f77b4
 ^           s         C1 #ff7f0e
 |           o         C2 #2ca02c
 v           d         C3 #d62728
Small        v         C4 #d9467bd

Color are robust on both, screen, print, projector.

# TODO:
 - define color list for colorblindness
 - make line style list
 - make plotstyle for including color bars.
    This should ONLY take up real estate on the right hand side of the figure
 - make template for different figures stacked on top of each other
 - legend.handlelength: check more line styles.
"""

import matplotlib
import matplotlib.pyplot as plt
import numpy as np
from matplotlib import ticker

# For 5 symbols in a plot
symbol_list_5 = ["^", "s", "o", "d", "v"]

# For 4 symbols in a plot
symbol_list_4 = ["^", "s", "d", "v"]

# 3 symbols in a plot
symbol_list_3 = ["^", "o", "v"]

# 2 symbols in a plot
symbol_list_2 = ["^", "v"]

# 1 symbol
symbol_list_1 = ["o"]


golden_ratio = 0.5 * (1.0 + 5**0.5)


def set_rcparams_dynamo(
    myParams: matplotlib.RcParams, num_cols: int = 1, ls: str = "thin"
) -> list[float]:
    """Half column width figures for revtex.

    See http://publishing.aip.org/authors/preparing-graphics

    Parameters
    ----------
    myParams: matplotlib.rcParams
        The parameters from matplotlib.rcParams
    num_cols: int
        number of columns the figure spans.
        1 column  = 3.37" by 2.08"
        2 columns = 6.69" by 4.13"
    ls: str
        Either "thick" or "thin". Defaults to "thin"

    Returns
    -------
    list[float]
        List of length 4: x position, y position, width, height
    """

    fig_dpi = 300.0
    fontsize = 8

    assert ls in {"thick", "thin"}
    linewidth = 0.75
    if ls == "thick":
        linewidth *= 2

    assert num_cols in {1, 2}

    # Define axis size to be used
    if num_cols == 1:
        ax_x0, ax_y0 = 0.2, 0.2
        axes_size = [ax_x0, ax_y0, 0.95 - ax_x0, 0.95 - ax_y0]
    elif num_cols == 2:
        ax_x0, ax_y0 = 0.1, 0.2
        axes_size = [ax_x0, ax_y0, 0.975 - ax_x0, 0.95 - ax_y0]

    # Figure size in inch
    fig_width_in = num_cols * 3.37
    fig_height_in = 3.37 / golden_ratio

    # Figure size and dpi
    myParams["figure.dpi"] = fig_dpi
    myParams["figure.figsize"] = [fig_width_in, fig_height_in]
    myParams["savefig.dpi"] = fig_dpi

    # Figure legend
    myParams["legend.framealpha"] = 1.0
    myParams["legend.fancybox"] = False
    myParams["legend.edgecolor"] = "k"
    myParams["patch.linewidth"] = 0.5  # For legend box borders
    myParams["legend.handlelength"] = 1.45  # Show nice, even
    # numbers for different line styles

    # Font and text
    myParams["text.usetex"] = True
    myParams["pdf.fonttype"] = 42
    myParams["font.family"] = "Times"
    myParams["font.size"] = fontsize
    myParams["axes.labelsize"] = fontsize
    myParams["legend.fontsize"] = fontsize

    # Line size and marker size
    myParams["lines.markersize"] = 3.0 * linewidth
    myParams["lines.linewidth"] = linewidth

    # Axes thickness
    myParams["axes.linewidth"] = 0.5
    # Enable minor ticks
    myParams["ytick.minor.visible"] = True
    myParams["xtick.minor.visible"] = True
    # Default ticks on both sides of the axes
    myParams["xtick.top"] = True
    myParams["xtick.bottom"] = True
    myParams["ytick.left"] = True
    myParams["ytick.right"] = True
    # All ticks point inward
    myParams["xtick.direction"] = "in"
    myParams["ytick.direction"] = "in"

    return axes_size


def log_tick_format(axes: plt.Axes, which: str, base: float = 10) -> plt.Axes:
    """Change the tick formatter.

    Change the logarithmic axes to use '1' and '10' (or the given base), i.e. without
    power, otherwise use the base to some power. For more robust and less error prone
    results, the plotting type is also re-set with the same base ('loglog', 'semilogx' and
    'semilogy').

    Modified from:
    https://stackoverflow.com/questions/21920233/matplotlib-log-scale-tick-label-number-formatting/33213196

    Parameters
    ----------
    axes: plt.Axes
        An axes object
    which: str
        Whether to update both x and y axis, or just one of them.
    base: float
        The base of the logarithm. Defaults to base 10 (same as loglog, etc.)

    Returns
    -------
    plt.Axes
        The updated axes object.

    Raises
    ------
    ValueError
        If the axis given to `which` is not recognized.
    """
    if which == "both":
        axs, pltype = ["xaxis", "yaxis"], "loglog"
    elif which == "x":
        axs, pltype = ["xaxis"], "semilogx"
    elif which == "y":
        axs, pltype = ["yaxis"], "semilogy"
    else:
        raise ValueError(
            "No valid axis found. 'which' must be either of 'both', 'x' or 'y'."
        )
    getattr(axes, pltype)(base=base)
    for ax in axs:
        f = getattr(axes, ax)
        f.set_major_formatter(
            ticker.FuncFormatter(
                lambda x, _: (
                    "{:g}".format(x)
                    if np.log(x) / np.log(base) in [0, 1]
                    else r"$"
                    + str(base)
                    + "^{"
                    + "{:g}".format(np.log(x) / np.log(base))
                    + r"}$"
                )
            )
        )
    return axes



"""Load and view the generated forcing.

This script will create a plot of the forcing, i.e. total emission vs time. The last saved
file is used unless a file is specified explicitly.

Providing a custom file means you can view any forcing so long as it is saved in the same
format that the synthetically created files are.
"""

import datetime
import os
import sys
from typing import List, Optional, Tuple

import cftime
import cosmoplots
import matplotlib.pyplot as plt
import numpy as np
import xarray as xr

import volcano_cooking.helper_scripts.functions as fnc


def view_forcing(
    ext: Optional[str] = None,
    in_file: Optional[str] = None,
    width: int = 2,
    style: str = "connected",
    dark: bool = False,
    save: bool = False,
    return_fig: bool = False,
) -> Optional[plt.Figure]:
    """View the forcing found in the generated total emission against time.

    Parameters
    ----------
    ext: str
        Extension of the file type used. Valid values are 'npz' and 'nc'.
    in_file: str, optional
        Full path (absolute or relative to where the function i called) and file name of
        a custom file to be used.
    width: int
        Width of the plot in number of columns. Default is 2.
    style: str
        Style of the plot. Default is 'connected'.
    dark: bool
        Plot with dark background. Defaults to False
    save: bool
        Save the plot. Defaults to False.
    return_fig: bool
        Return the figure. Defaults to False.

    Returns
    -------
    plt.Figure
        The figure of the plot.
    """
    _FIG_STD_ = cosmoplots.set_rcparams_dynamo(plt.rcParams, num_cols=width)
    t, tes = frc_datetime2float(ext, in_file)
    if dark:
        plt.style.use("dark_background")
    fig = plt.figure()
    ax = fig.add_axes(_FIG_STD_)
    if style == "connected":
        ax.plot(t, tes, "+-", color="r")
    elif style == "bars":
        bar_width = (t[-1] - t[0]) * 0.005
        ax.bar(t, tes, color="r", width=bar_width)
    else:
        ax.plot(t, tes, "+-", color="r")
    plt.xlabel("Time")
    plt.ylabel("Total Emission")
    if save:
        filename = check_dir(".png")
        plt.savefig(filename, format="png")
        print(f"Saved to: {os.path.join(os.getcwd(), filename)}")
    if return_fig:
        return plt.gcf()
    plt.show()
    return None


def frc_datetime2float(
    ext: Optional[str] = None, in_file: Optional[str] = None
) -> Tuple[np.ndarray, np.ndarray]:
    """Convert time axis from datetime to float.

    Parameters
    ----------
    ext: str
        Extension of the file type used. Valid values are 'npz' and 'nc'.
    in_file: str, optional
        Full path (absolute or relative to where the function i called) and file name of a
        custom file to be used.

    Returns
    -------
    np.ndarray:
        Time axis of the input file
    np.ndarray:
        Values of the input file

    Raises
    ------
    ValueError
        If the date created fails to be formatted from ints to string and datetime objects
    """
    if ext is None and in_file is None:
        yoes, moes, does, tes = load_forcing(".nc", in_file=in_file)
    elif isinstance(ext, str):
        yoes, moes, does, tes = load_forcing(ext, in_file=in_file)
    elif isinstance(in_file, str) and in_file.endswith((".nc", ".npz")):
        ext = in_file.split(".")[-1]
        yoes, moes, does, tes = load_forcing(ext, in_file=in_file)
    else:
        raise ValueError(f"This file cannot be viewed: {in_file} = ")

    dates: List[datetime.datetime] = []
    dates_ap: List[str] = []
    d_app = dates.append
    d_ap_app = dates_ap.append
    for y, m, d in zip(yoes, moes, does):
        y = str("0" * (4 - len(str(y)))) + f"{y}"
        m = str("0" * (2 - len(str(m)))) + f"{m}"
        d = str("0" * (2 - len(str(d)))) + f"{d}"
        if len(y) != 4 or len(m) != len(d) != 2:
            raise ValueError(
                "Year, month and/or day is not formatted properly. "
                + f"{len(y)} = , {len(m)} = , {len(d)} = , need 4, 2, 2."
            )
        d_app(datetime.datetime.strptime(f"{y}{m}{d}", "%Y%m%d"))
        d_ap_app(f"{y}-{m}-{d}")

    shift = dates_ap[0][:4]
    t_0 = f"{shift}-01-01 00:00:00"
    t = (
        cftime.date2num(
            dates, f"days since {t_0}", calendar="noleap", has_year_zero=True
        )
        + float(shift) * 365
    ) / 365
    return t, tes


def check_dir(ext: str) -> str:
    """Check if the directory exist and there is no other file with the same name.

    Parameters
    ----------
    ext: str
        The file ending

    Returns
    -------
    str
        The path and name of the file that can be created
    """
    if ext[0] == ".":
        ext = ext[1:]
    d = os.path.join("data", "output")
    if not os.path.isdir(d):
        os.makedirs(d)
    now = datetime.datetime.now().strftime("%Y%m%d_%H%M")
    out_file = os.path.join(d, f"synthetic_volcanoes_{now}.{ext}")
    if os.path.isfile(out_file):
        sys.exit(f"The file {out_file} already exists.")
    return out_file


def load_forcing(
    ext: str, in_file: Optional[str] = None
) -> Tuple[np.ndarray, np.ndarray, np.ndarray, np.ndarray]:
    """Load variables from the last saved file of given extension.

    Data about the date and total emission found in either the last npz or nc file.

    Parameters
    ----------
    ext: str
        Extension of the file that should be used
    in_file: str, optional
        Full path (absolute or relative to where the function i called) and file name of a
        custom file to be used.

    Returns
    -------
    np.ndarray, np.ndarray, np.ndarray, np.ndarray
        Dates and forcing (total emissions) is returnd in the order years, months, days,
        total emission

    Raises
    ------
    NameError
        If the extension is not either 'npz' or 'nc', no files can be found and the
        variables 'yoes', 'moes', 'does' and 'tes' cannot be found.
    """
    if in_file is None:
        file = fnc.find_last_output(ext)
    else:
        file = fnc.find_file(in_file)
    if "npz" in ext:
        with np.load(file, "r", allow_pickle=True) as f:
            yoes = f["yoes"]
            moes = f["moes"]
            does = f["does"]
            tes = f["tes"]
            f.close()
    elif "nc" in ext:
        f = xr.open_dataset(file, decode_times=False)
        yoes = f.variables["Year_of_Emission"].data
        moes = f.variables["Month_of_Emission"].data
        does = f.variables["Day_of_Emission"].data
        tes = f.variables["Total_Emission"].data
    else:
        raise NameError(
            "Names 'yoes', 'moes', 'does' and 'tes' are not found. "
            + f"Extension was {ext} = , but shuld be 'npz' or 'nc'."
        )

    return yoes, moes, does, tes


def main():
    # Remember to remove plt.show() when timing
    # import timeit
    # t1 = timeit.timeit(lambda: view_forcing("nc"), number=1000)
    # t2 = timeit.timeit(lambda: view_forcing("npz"), number=1000)
    # print(f"{t1 = }, {t2 = }")
    view_forcing("npz")  # Twice as fast as nc


if __name__ == "__main__":
    main()



"""Script that implement the class which correctly created a netCDF4 file and saves it."""

import datetime
import json
import os
import sys
from abc import ABC, abstractmethod
from typing import List, Optional

import numpy as np
import xarray as xr

import volcano_cooking.modules.convert as convert
import volcano_cooking.modules.create as create


class Data:
    """Create data for a volcanic forcing file.

    Attributes
    ----------
    make_dataset: method
        Create the xarray data set with the numpy arrays provided to the class on
        initialisation.
    save_to_file: method
        Save the generated xarray data set as a netCDF4 file, and the defining numpy
        arrays to a .npz file.
    """

    def __init__(
        self,
        eruptions: np.ndarray,
        yoes: np.ndarray,
        moes: np.ndarray,
        does: np.ndarray,
        lats: np.ndarray,
        lons: np.ndarray,
        tes: np.ndarray,
        veis: np.ndarray,
        miihs: np.ndarray,
        mxihs: np.ndarray,
    ) -> None:
        """Initialise object by giving all variables that should be put in the .nc file.

        Parameters
        ----------
        eruptions: np.ndarray
            Eruption number
        yoes: np.ndarray
            Years (four digit number) that the volcanoes occurred
        moes: np.ndarray
            Months (two digit number) that the volcanoes occurred
        does: np.ndarray
            Days (two digit number) that the volcanoes occurred
        lats: np.ndarray
            Latitudinal coordinate
        lons: np.ndarray
            Longitudinal coordinate
        tes: np.ndarray
            Total emission from a volcano at corresponding date
        veis: np.ndarray
            VEI (Volcanic Explosivity Index) of a volcano at corresponding date
        miihs: np.ndarray
            Minimum injection height
        mxihs: np.ndarray
            Maximum injection height
        """
        self.eruptions = eruptions
        self.yoes = yoes
        self.moes = moes
        self.does = does
        self.lats = lats
        self.lons = lons
        self.tes = tes
        self.veis = veis
        self.miihs = miihs
        self.mxihs = mxihs
        self.__run_check()
        self.my_frc: xr.Dataset

    def __run_check(self) -> None:
        """Check the data type of each numpy array.

        The original netCDF file that is re-produced has specific data types on all data
        arrays. This method checks to see if the types are set correctly.

        Raises
        ------
        ValueError
            If any of the ten numpy arrays contain data of wrong type a ValueError is
            raised. ValueError is also raised if they do not have the same shape.
        """
        stack = [
            self.eruptions.shape,
            self.yoes.shape,
            self.moes.shape,
            self.does.shape,
            self.lats.shape,
            self.lons.shape,
            self.tes.shape,
            self.veis.shape,
            self.miihs.shape,
            self.mxihs.shape,
        ]
        if not (np.diff(np.vstack(stack).reshape(len(stack), -1), axis=0) == 0).all():
            raise ValueError(f"The arrays have different shapes:\n{stack}")
        else:
            del stack
        if (
            self.eruptions.dtype != np.int8
            or self.veis.dtype != np.int8
            or self.moes.dtype != np.int8
            or self.does.dtype != np.int8
        ):
            raise ValueError(
                f"self.eruptions.dtype = {self.eruptions.dtype}, "
                + f"self.veis.dtype = {self.veis.dtype}, "
                + f"self.moes.dtype = {self.moes.dtype}, "
                + f"self.does.dtype = {self.does.dtype}. All must be int8."
            )
        if self.yoes.dtype != np.int16:
            raise ValueError(f"{self.yoes.dtype} = . Need int16.")
        if (
            self.tes.dtype != np.float32
            or self.mxihs.dtype != np.float32
            or self.miihs.dtype != np.float32
            or self.lats.dtype != np.float32
            or self.lons.dtype != np.float32
        ):
            raise ValueError(
                f"self.tes.dtype = {self.tes.dtype}, "
                + f"self.mxihs.dtype = {self.mxihs.dtype}, "
                + f"self.miihs.dtype = {self.miihs.dtype}, "
                + f"self.lats.dtype = {self.lats.dtype}"
                + f"self.lons.dtype = {self.lons.dtype}. All must be float32."
            )

    def make_dataset(self) -> None:
        """Make an xarray Dataset object where all variables are stored."""
        self.my_frc = xr.Dataset(
            data_vars=dict(
                Eruption=(["Eruption_Number"], self.eruptions),
                VEI=(["Eruption_Number"], self.veis),
                Year_of_Emission=(["Eruption_Number"], self.yoes),
                Month_of_Emission=(["Eruption_Number"], self.moes),
                Day_of_Emission=(["Eruption_Number"], self.does),
                Latitude=(["Eruption_Number"], self.lats),
                Longitude=(["Eruption_Number"], self.lons),
                Total_Emission=(["Eruption_Number"], self.tes),
                Maximum_Injection_Height=(["Eruption_Number"], self.mxihs),
                Minimum_Injection_Height=(["Eruption_Number"], self.miihs),
            ),
            # Global attributes
            attrs=dict(
                Creator="Eirik R. Enger",
                DOI="#####",
                Citation="#####",
                Notes="Emissions source file created with the `volcano-cooking` python package.",
            ),
        )

        size = len(self.yoes)
        # Names are unimportant. The real data set would list the name of the volcano
        # here, e.g. Mt. Pinatubo.
        names = "".join("N, " for _ in range(size))
        names = names[:-2]  # Removing the last comma and whitespace
        self.my_frc.Longitude.encoding["_FillValue"] = False
        self.my_frc["Eruption"] = self.my_frc.Eruption.assign_attrs(
            # The volcano number is a reference to the geographical location of the
            # eruption. The volcano name is its name and the note/reference is where in
            # the literature you can find the values related to the eruption. All this is
            # unimportant given that this is just made up data.
            Volcano_Number=np.ones(size) * 123456,
            Volcano_Name=names,
            Notes_and_References=names,
        )
        self.my_frc["VEI"] = self.my_frc.VEI.assign_attrs(
            Notes="Volcanic_Explosivity_Index_based_on_Global_Volcanism_Program"
        )
        self.my_frc["Latitude"] = self.my_frc.Latitude.assign_attrs(Units="-90_to_+90")
        self.my_frc["Longitude"] = self.my_frc.Longitude.assign_attrs(
            Units="Degrees_East"
        )
        self.my_frc["Total_Emission"] = self.my_frc.Total_Emission.assign_attrs(
            Units="Tg_of_SO2"
        )
        self.my_frc["Maximum_Injection_Height"] = (
            self.my_frc.Maximum_Injection_Height.assign_attrs(
                Units="km_above_mean_sea_level"
            )
        )
        self.my_frc["Minimum_Injection_Height"] = (
            self.my_frc.Minimum_Injection_Height.assign_attrs(
                Units="km_above_mean_sea_level"
            )
        )

    def save_to_file(self) -> None:
        """Save the xarray Dataset object to a .nc file using the netCDF4 format.

        For easier access to the forcing data, dates and total emissions are also saved to
        a .npz file.
        """
        self.__save_to_nc_file()
        self.__save_to_npz_file()

    def __save_to_npz_file(self) -> None:
        """Save the original forcing data to a .npz file."""
        out_file = self.check_dir("npz")
        np.savez(out_file, yoes=self.yoes, moes=self.moes, does=self.does, tes=self.tes)

    def __save_to_nc_file(self) -> None:
        """Save the xarray Dataset object to a .nc file using the netCDF4 format.

        Raises
        ------
        ValueError
            If the attribute `my_frc` has not been filled yet, `ValueError` is raised
            telling the user to first call the `make_dataset` method.
        """
        if not hasattr(self, "my_frc"):
            raise ValueError("You must make the dataset with 'make_dataset' first.")
        out_file = self.check_dir("nc")
        # The format is important for when you give the .nc file to the .ncl script that
        # creates the final forcing file.
        self.my_frc.to_netcdf(out_file, "w", format="NETCDF4")
        # self.my_frc.to_netcdf(out_file, "w", format="NETCDF4_CLASSIC")

    @staticmethod
    def check_dir(ext: str, name: str = "synthetic_volcanoes") -> str:
        """Check if the directory exist and there is no other file with the same name.

        Parameters
        ----------
        ext: str
            The file ending
        name: str
            The file name. Defaults to 'synthetic_volcanoes'.

        Returns
        -------
        str
            The path and name of the file that can be created
        """
        if ext[0] == ".":
            ext = ext[1:]
        d = os.path.join("data", "output")
        if not os.path.isdir(d):
            os.makedirs(d)
        now = datetime.datetime.now().strftime("%Y%m%d_%H%M%S")
        out_file = os.path.join(d, f"{name}_{now}.{ext}")
        if os.path.isfile(out_file):
            sys.exit(f"The file {out_file} already exists.")
        return out_file


class Generate(ABC):
    """ABC for generating data with different properties."""

    def __init__(self, size: int, init_year: int, file: Optional[str] = None) -> None:
        """Initialise object by giving all variables that should be put in the .nc file.

        Parameters
        ----------
        size: int
            The total number of volcanoes
        init_year: int
            The first possible year for a volcanic eruption
        file: str, optional
            Path to the json file.
        """
        self.eruptions: np.ndarray
        self.yoes: np.ndarray
        self.moes: np.ndarray
        self.does: np.ndarray
        self.lats: np.ndarray
        self.lons: np.ndarray
        self.tes: np.ndarray
        self.veis: np.ndarray
        self.miihs: np.ndarray
        self.mxihs: np.ndarray
        self.size = size
        self.init_year = init_year
        self.file = file

    @abstractmethod
    def gen_dates_totalemission_vei(self) -> None:
        """Generate dates, total emission and VEI."""

    def gen_lat_lon(self) -> None:
        """Generate latitude and longitude.

        If not set by the subclasses, a fixed location of 0deg N, 1deg E will be used.
        """
        if not hasattr(self, "lats"):
            # Place all volcanoes at equator
            self.lats = np.zeros(self.size, dtype=np.float32)  # Equator
        if not hasattr(self, "lons"):
            self.lons = np.ones(self.size, dtype=np.float32)

    def gen_injection_heights(self) -> None:
        """Generate minimum and maximum injection heights.

        By default, the minimum and maximum heights of injection which are both assumed
        functions of `veis`, the Volcanic Explosivity Index. If only one of them is
        given, the other is given a default constant value of 18 (20) for the minimum
        (maximum) injection height.
        """
        if not hasattr(self, "miihs") and not hasattr(self, "mxihs"):
            self.miihs, self.mxihs = convert.vei_to_injectionheights(self.veis)
            return
        if not hasattr(self, "miihs"):
            self.miihs = np.ones(self.size, dtype=np.float32) * 18
            return
        if not hasattr(self, "mxihs"):
            self.mxihs = np.ones(self.size, dtype=np.float32) * 20
            return

    def _gen_rest(self) -> None:
        """Generate the rest with default settings.

        The rest refer to `eruptions`, the eruption number. These do not matter when
        working with synthetic data.
        """
        # This only (in the real data set) store a number for each volcanic event,
        # increasing as new events occur. If a volcanic eruption have several emissions
        # listed in the forcing file the number is repeated, giving a list similar to
        # [1, 2, 3, 4, 4, 4, 5, 5, 6, 7, ...].  Here, the fourth and fifth eruptions
        # lasted long enough and to get more samples in the forcing file. Anyway, its
        # most likely not important, so I just put gibberish in it.
        self.eruptions = np.random.randint(1, high=20, size=self.size, dtype=np.int8)

    def generate(self) -> None:
        """Generate all data arrays needed in the volcanic forcing netCDF file.

        Ten arrays of equal shape are created; eruption number, year-month-day,
        latitude-longitude, total emission, volcanic explosivity index and minimum &
        maximum injection heights.
        """
        self.gen_dates_totalemission_vei()
        if len(self.veis) != self.size:
            self.size = len(self.veis)
        self.gen_lat_lon()
        self.gen_injection_heights()
        # Make sure minimum heights are actually smaller than maximum heights
        for idx, (i, j) in enumerate(zip(self.miihs, self.mxihs)):
            self.miihs[idx] = min(i, j)
            self.mxihs[idx] = max(i, j)
        self.mxihs = self.mxihs.astype(np.float32)
        self.miihs = self.miihs.astype(np.float32)
        if self.file is None:
            # Shift first eruption to occur before initial year
            self.yoes -= abs(self.init_year - self.yoes[0]) + 1
        self._gen_rest()

    def get_arrays(self) -> List[np.ndarray]:
        """Return all generated data.

        Returns
        -------
        list
            A list of all arrays. Order is the same as the input to the Data class

        Raises
        ------
        AttributeError
            If any one array has not been set, i.e. is still a 'type' object
        """
        arrs = [
            self.eruptions,
            self.yoes,
            self.moes,
            self.does,
            self.lats,
            self.lons,
            self.tes,
            self.veis,
            self.miihs,
            self.mxihs,
        ]
        if any(len(a) == 0 for a in arrs):
            raise AttributeError(
                "Attribute(s) do not exist. Run the 'generate' method."
            )
        return arrs


class GenerateRandomNormal(Generate):
    def gen_dates_totalemission_vei(self):
        self.yoes, self.moes, self.does = create.random_dates(self.size, self.init_year)
        # We don't want eruptions that have a VEI greater than 6.
        self.veis = np.random.normal(4, 1, size=self.size).round().astype(np.int8) % 7
        self.tes = convert.vei_to_totalemission(self.veis)


class GenerateFPP(Generate):
    def gen_dates_totalemission_vei(self):
        self.yoes, self.moes, self.does, self.tes = create.fpp_dates_and_emissions(
            self.size, self.init_year
        )
        self.veis = convert.totalemission_to_vei(self.tes)


class GenerateRegularIntervals(Generate):
    def gen_dates_totalemission_vei(self) -> None:
        self.yoes, self.moes, self.does, self.tes = create.regular_intervals()
        self.veis = convert.totalemission_to_vei(self.tes)


class GenerateSingleVolcano(Generate):
    def gen_dates_totalemission_vei(self) -> None:
        self.yoes, self.moes, self.does, self.tes = create.single_date_and_emission(
            self.init_year
        )
        self.veis = convert.totalemission_to_vei(self.tes)


class GenerateFromFile(Generate):
    def gen_dates_totalemission_vei(self) -> None:
        if self.file is None:
            raise AttributeError("Cannot use GenerateFromFile without a json file.")
        with open(self.file) as f:
            a = json.load(f)
        self = create.from_json(a, self)
        self.veis = convert.totalemission_to_vei(self.tes)



"""CLI to interpolate raw volcanic forcing to npz format.

It takes a filename as a mandatory input parameter.
"""

import os
from typing import List, Optional, Tuple

import click
import numpy as np

import volcano_cooking.helper_scripts as hs


def sparse_to_lin(
    t: np.ndarray,
    sy: Optional[int] = None,
    ey: Optional[int] = None,
    samples_per_year: int = 12,
    remove: int = 1,
) -> Tuple[np.ndarray, List, List]:
    """Re-sample an uneven time axis to one with linear spacing.

    The function returns the new linearly spaced time axis, as well as two
    masks, one for the original time and one for the new time, which makes them
    as equal as possible for the given resolution (resolution is set to 12,
    i.e. months).

    Parameters
    ----------
    t: np.ndarray
        The time axis with uneven time stamps.
    sy: int, optional
        The assumed first year of the time axis
    ey: int, optional
        The assumed end year of the time axis
    samples_per_year: int
        The number of samples per year.
    remove: int
        Remove the number of elements at the end of the time array equal to 'remove'.

    Returns
    -------
    np.ndarray:
        The new linearly spaced time axis
    list:
        Masking of the original time axis that give the time stamps we want to
        keep that match best with the new time axis
    list:
        Masking of the new time axis that give the time stamps kept from the original

    Examples
    --------
    >>> t = np.array(
    >>>     [
    >>>         0, 1e-3, 2e-3, 3e-3, 4e-3, 5e-3, 4, 6, 7.85, 7.91,
    >>>         7.92, 7.925, 7.93, 7.96, 7.99, 8.001, 8.002, 8.003,
    >>>     ]
    >>> )
    >>> t2, t_m, t2_m = sparse_to_lin(t)
    >>> assert t[t_m].shape == t2[t2_m].shape
    Elements are within half a time step from each other: True
    """
    # New time axis (stop at December)
    sy = int(t[0]) if sy is None else sy
    ey = int(t[-1] + 1) if ey is None else ey
    t_i = np.linspace(sy, ey, (ey - sy) * samples_per_year + 1)[
        : -int(remove * samples_per_year / 12)
    ]
    # Find indices in `t_i` that are closest to `t`. If one element in `t_i` is closest
    # to two or more elements in `t`, only the one closest in `t` is kept (avoid
    # repetition of indices).
    # Make `t` a row vector (1, N) and `t_i` a column vector (M, 1), compute the
    # absolute difference -> (M, N) and pick the minimum along N -> (N,)
    # `mask` represents which index of the new linspaced time array the given element of
    # the original time array is closest. `mask_idx` represents unique elements of the
    # new time array that should be used, and `mask_sections` represents the cluster of
    # elements in the original that are closest to a given element of the new linspaced
    # time array.
    mask = np.abs(t[None, :] - t_i[:, None]).argmin(axis=0)
    _, mask_idx = np.unique(mask, return_index=True)
    if len(t) == 1:
        mask_sections = [mask[mask_idx[-1] :]]
    else:
        mask_sections = [mask[m : mask_idx[i + 1]] for i, m in enumerate(mask_idx[:-1])]
        if not mask_sections:
            return t_i, [], []
        mask_sections.append(mask[mask_idx[-1] :])
    # Now we find the indices of `t` that we want to keep. `c` works as the index of the
    # first item in each array inside mask_sections
    c = 0
    t_mask: List[int] = []
    t_mask_app = t_mask.append
    ti_mask: List[int] = []
    ti_mask_app = ti_mask.append
    for arr in mask_sections:
        # Find index of t (arr) that is closest to t_i[arr]
        idx = abs(t[c : c + len(arr)] - t_i[arr[0]]).argmin()
        t_mask_app(c + idx)
        ti_mask_app(arr[0])
        c += len(arr)
    closeness = np.allclose(t[t_mask], t_i[ti_mask], atol=1 / (2 * samples_per_year))
    print(f"Elements are within half a time step from each other: {closeness}")
    return t_i, t_mask, ti_mask


@click.command()
@click.argument("filename", type=click.Path(exists=True), required=True)
@click.option(
    "--save/--no-save",
    "-s/-S",
    default=False,
    show_default=True,
    type=bool,
    help="Save the new synthetic forcing array.",
)
@click.option(
    "-sy",
    default=None,
    show_default=True,
    type=int,
    help="Set the starting year.",
)
@click.option(
    "-ey",
    default=None,
    show_default=True,
    type=int,
    help="Set the end year.",
)
@click.option(
    "-spy",
    "--samples-per-year",
    "samples_per_year",
    default=12,
    show_default=True,
    type=int,
    help="Set the number of samples per year.",
)
@click.option(
    "-lm",
    default="Dec",
    show_default=True,
    type=str,
    help="Set the last month using the three first letters.",
)
@click.option(
    "--show/--no-show",
    default=False,
    show_default=True,
    type=bool,
    help="Show an example plot of original and new.",
)
def main(
    filename: str,
    save: bool,
    show: bool,
    sy: int,
    ey: int,
    samples_per_year: int,
    lm: str,
):
    """View a plot of `filename`."""
    # Decide on the last month
    lm = lm.lower()
    # fmt: off
    months = {
        "jan": 1, "feb": 2, "mar": 3, "apr": 4, "may": 5, "jun": 6,
        "jul": 7, "aug": 8, "sep": 9, "oct": 10, "nov": 11, "dec": 12,
    }
    # fmt: on
    if lm not in months:
        print(f"{lm} =  is not a valid month. Using 'Dec'...")
        last_month = 1
    else:
        last_month = 13 - months[lm]
    # Grab forcing file with list of timing of volcanic events (uneven in time)
    t, tes = hs.frc_datetime2float(in_file=filename)
    # Remove data points outside of `sy` and `ey`
    if sy is not None:
        tes = tes[t >= sy]
        t = t[t >= sy]
    if ey is not None:
        tes = tes[t <= ey]
        t = t[t <= ey]
    # Get a linspaced time axis and two masks.
    t_i, t_mask, ti_mask = sparse_to_lin(t, sy, ey, samples_per_year, last_month)

    # Create new total emission array with linear sampling in time
    tes_new = np.zeros_like(t_i)
    tes_new[ti_mask] = tes[t_mask]
    if save:
        filebase = os.path.splitext(filename)[0]
        np.savez(f"{filebase}-linspace.npz", times=t_i, data=tes_new)
    if show:
        import matplotlib.pyplot as plt

        plt.plot(t_i, tes_new, "x-", label="Zeros")
        plt.plot(t, tes, "+-", label="Orig")
        plt.plot(t_i[ti_mask], tes[t_mask], "*-r", label="New")
        plt.legend()
        plt.grid()
        plt.show()


def example() -> None:
    """Show an example use of function."""
    # fmt: off
    a = np.array(
        [
            0, 1e-3, 2e-3, 3e-3, 4e-3, 5e-3, 4, 6, 7.85, 7.91,
            7.92, 7.925, 7.93, 7.96, 7.99, 8.001, 8.002, 8.003,
        ]
    )
    # fmt: on
    sparse_to_lin(a)


if __name__ == "__main__":
    main()



"""Adjust emissions and emission heights to go into the CESM2.

Too much aerosols in a column can trigger a warning saying that the `Aerosol optical depth
is unreasonably high in this layer.` We therefore restrict this, as well as correcting for
the injections heights, making sure they are within the range of the model layers.

Note
----
Modified from script provided by Herman Fæhn Fuglestvedt and using
http://svn.code.sf.net/p/codescripts/code/trunk/ncl/emission/createVolcEruptV3.ncl
"""

import os
from typing import Tuple

import numpy as np
import xarray as xr


def adjust_altitude_range(
    miihs: np.ndarray, mxihs: np.ndarray, tes: np.ndarray
) -> Tuple[np.ndarray, np.ndarray]:
    """Adjust altitude range to within the layers of CESM2.

    Parameters
    ----------
    miihs: np.ndarray
        Minimum injection heights
    mxihs: np.ndarray
        Maximum injection heights
    tes: np.ndarray
        Total emissions

    Returns
    -------
    Tuple[np.ndarray, np.ndarray]
        Adjusted minimum and maximum injection heights

    Note
    ----
    Python implementation of the correction from:
    http://svn.code.sf.net/p/codescripts/code/trunk/ncl/emission/createVolcEruptV3.ncl
    """
    threshold = 3.5  # Tg
    idx = (tes > threshold) & (mxihs > 20)
    mxihs[idx] = 20
    # If some of the lower boundaries at `idx` are higher than 18, we set them to 18.
    miihs[idx & (miihs > 18)] = 18
    return miihs, mxihs


def adjust_emissions(
    miihs: np.ndarray,
    mxihs: np.ndarray,
    tes: np.ndarray,
    e_lons: np.ndarray,
    e_lats: np.ndarray,
) -> np.ndarray:
    """Calculate emissions that are accepted by the CESM2.

    Too much aerosols in a column can trigger a warning saying that the `Aerosol optical
    depth is unreasonably high in this layer`. We therefore restrict this, should be on
    the order of about `1e10`.

    Parameters
    ----------
    miihs: np.ndarray
        Minimum injection heights
    mxihs: np.ndarray
        Maximum injection heights
    tes: np.ndarray
        Total emissions
    e_lons: np.ndarray
        Longitude of the eruption
    e_lats: np.ndarray
        Latitude of the eruption

    Returns
    -------
    np.ndarray
        The total column emission

    Note
    ----
    Modified from script provided by Herman Fæhn Fuglestvedt and according to
    http://svn.code.sf.net/p/codescripts/code/trunk/ncl/emission/createVolcEruptV3.ncl

    Raises
    ------
    FileNotFoundError
        If the coordinates file cannot be found.
    """
    file = os.path.join(
        "data",
        "originals",
        "fv_1.9x2.5_L30",
    )
    if not os.path.exists(f"{file}.nc"):
        raise FileNotFoundError(f"{file} not found. Consult README on how to download.")
    mass_threshold = 15
    idx = tes > mass_threshold
    mass_factor = 1 / 1.8  # scaling factor for large eruptions
    tes[idx] *= mass_factor

    duration = 21600  # 6 hours in seconds
    avogadros_number = 6.022140857e23  # Avogadro's constant in "mol^{-1}"
    d_earth = 12742e3  # in "m"
    s_earth = np.pi * d_earth**2
    # dalt = 1e3
    dalt = mxihs * 1e5 - miihs * 1e5  # Emission depth in cm
    m3_to_cm3 = 1.0e6
    kg2g = 1000
    m_mass, fraction = (64.06, 1.0)

    grid = xr.open_dataset(os.path.join("data", "originals", "fv_1.9x2.5_L30.nc"))
    gw = grid["gw"]  # Latitude/Gauss weights
    nLon = len(grid["lon"])
    # nLat = len(grid["lat"])
    lat_data = grid["lat"]
    column_area = s_earth / nLon * gw / np.sum(gw)  # in "m"
    idx = (np.abs(e_lats - lat_data)).argmin()
    area = column_area[idx] * len(e_lons)
    volume = area.data * dalt  # volume of one level
    v_cm3 = volume * m3_to_cm3  # Volume in cm3
    # tot_molecules = tes * kg2g / m_mass * avogadros_number  # "imass" is in "kg"
    # emis = tot_molecules / (v_cm3 * duration)
    # total = ((v_cm3 * emis * duration) / avogadros_number * m_mass) / kg2g * fraction
    # print(total)

    emis = tes / area.data / duration
    column_emission = emis * kg2g / m_mass * avogadros_number  # "imass" is in "kg"
    # emis_rate = column_emission / dalt
    _ = ((v_cm3 * emis * duration) / avogadros_number * m_mass) / kg2g * fraction
    # print(total)

    # stratvolc =

    return column_emission



"""Compare two variables across the original and synthetic forcing files."""

import matplotlib.pyplot as plt
import xarray as xr

import volcano_cooking.helper_scripts.functions as fnc


def x_vs_y() -> None:
    yax_file = fnc.find_original()
    xax_file = fnc.find_last_output("nc")

    f1 = xr.open_dataset(xax_file, decode_times=False)
    l1 = "Synthetic"
    f2 = xr.open_dataset(yax_file, decode_times=False)
    l2 = "Data set"

    var_list = [
        "Eruption",
        "VEI",
        "Year_of_Emission",
        "Month_of_Emission",
        "Day_of_Emission",
        "Latitude",
        "Longitude",
        "Total_Emission",
        "Maximum_Injection_Height",
        "Minimum_Injection_Height",
    ]

    x_ax = 7
    y_ax = 8
    x1 = f1.variables[var_list[x_ax]].data
    y1 = f1.variables[var_list[y_ax]].data
    x2 = f2.variables[var_list[x_ax]].data
    y2 = abs(f2.variables[var_list[y_ax]].data)

    plt.figure(figsize=(12, 9))
    plt.loglog()
    shift = 0.05 if x_ax == 1 else 0
    plt.scatter(x1 - shift, y1, marker="d", label=l1)
    plt.scatter(x2 + shift, y2, marker="x", label=l2)
    plt.xlabel(var_list[x_ax])
    plt.ylabel(var_list[y_ax])
    plt.legend()
    plt.show()


def main():
    x_vs_y()


if __name__ == "__main__":
    main()



"""Script to put any function that creates correctly formatted dates."""

import datetime as dt
from itertools import cycle
from typing import List, Tuple, Union

import cftime
import numpy as np

from volcano_cooking.modules import create


def single_date_and_emission(
    init_year: int,
) -> Tuple[np.ndarray, np.ndarray, np.ndarray, np.ndarray]:
    """Create a single volcano three years after `init_year`.

    Parameters
    ----------
    init_year: int
        The first year dates should appear in

    Returns
    -------
    yoes: np.ndarray
        Array of length 'size' with the year of a date
    moes: np.ndarray
        Array of length 'size' with the month of a date
    does: np.ndarray
        Array of length 'size' with the day of a date
    veis: np.ndarray
        Array of length 'size' with the VEI as a 1D numpy array
    """
    yoes = np.array([init_year - 1, init_year + 2, init_year + 100], dtype=np.int16)
    moes = np.array([1, 1, 1], dtype=np.int8)
    does = np.array([15, 15, 15], dtype=np.int8)
    tes = np.array([1, 400, 1], dtype=np.float32)
    return yoes, moes, does, tes


def random_dates(
    size: int, init_year: Union[float, str]
) -> Tuple[np.ndarray, np.ndarray, np.ndarray]:
    """Create random dates and place them in order.

    The randomness do not follow a specific model. This is the simplest case where we just
    use the 'randint' function from 'numpy'.

    Parameters
    ----------
    size: int
        The total number of dates that should be made.
    init_year: float, str
        The first year dates should appear in

    Returns
    -------
    yoes: np.ndarray
        Array of length 'size' with the year of a date
    moes: np.ndarray
        Array of length 'size' with the month of a date
    does: np.ndarray
        Array of length 'size' with the day of a date

    Raises
    ------
    ValueError
        If `size` is non-positive.
    """
    if size < 1:
        raise ValueError(f"{size} = , but must be > 0.")
    yoes = np.cumsum(np.random.randint(0, high=2, size=size)).astype(np.int16) + int(
        init_year
    )
    moes = np.random.randint(1, high=12, size=size, dtype=np.int8)
    does = np.random.randint(1, high=28, size=size, dtype=np.int8)
    # Make sure dates are increasing!
    # idx = list of arrays of indexes pointing to the same year
    idxs = [np.where(yoes == i)[0] for i in np.unique(yoes)]
    # sort months with indexes
    # sort days with indexes
    for idx in idxs:
        moes[idx] = np.sort(moes[idx])
        does[idx] = np.sort(does[idx])

    return yoes, moes, does


def regular_intervals():
    """Create dates with regular intervals.

    This function creates dates with regular intervals and a pre-defined magnitude as
    VEI.

    Returns
    -------
    yoes: np.ndarray
        Array of length 'size' with the year of a date
    moes: np.ndarray
        Array of length 'size' with the month of a date
    does: np.ndarray
        Array of length 'size' with the day of a date
    veis: np.ndarray
        Array of length 'size' with the VEI
    """
    year_sep = 2
    init_year = 1850
    month = 3
    day = 15
    size = 100
    te = cycle([5, 1000, 40, 200])
    yoes = np.zeros(size, dtype=np.int16) + init_year
    moes = np.zeros(size, dtype=np.int8) + month
    does = np.zeros(size, dtype=np.int8) + day
    tes = np.ones(size, dtype=np.float32)
    for i in range(size):
        yoes[i] += i * year_sep
        tes[i] = next(te)
    return yoes, moes, does, tes


def fpp_dates_and_emissions(
    size: int, init_year: int
) -> Tuple[np.ndarray, np.ndarray, np.ndarray, np.ndarray]:
    """Create random ordered dates and total emissions.

    Data is created from an FPP process, specifically the arrival time and amplitude is
    used to set dates and total emissions.

    Parameters
    ----------
    size: int
        The total number of dates that should be made.
    init_year: int
        The first year dates should appear in

    Returns
    -------
    yoes: np.ndarray
        Array of length 'size' with the year of a date
    moes: np.ndarray
        Array of length 'size' with the month of a date
    does: np.ndarray
        Array of length 'size' with the day of a date
    tes:
        Array of length 'size' with the Total_Emission as a 1D numpy array
    """
    f = create.StdFrc(fs=12, total_pulses=size)
    while True:
        ta, amp = f.get_frc()
        # Can't have years beyond 9999
        if int(ta[-1]) + init_year > 9999:
            prev_size = int(ta[-1]) + init_year
            mask = np.argwhere(ta + init_year < 9999)
            ta = ta[mask].flatten()
            amp = amp[mask].flatten()
            size = len(amp)
            print(
                "Can't create dates for years beyond 9999. Keep this in mind when "
                + f"setting `init_year` and `size`. Size: {prev_size} -> {size}."
            )
        # Go from float to YYYY-MM-DD
        dates: List[cftime.datetime] = []
        dates_ap = dates.append
        for n in ta:
            result = cftime.datetime(
                int(n) + init_year, 1, 1, calendar="noleap"
            ) + dt.timedelta(days=(n % 1) * 365)
            dates_ap(result)
        # Dates should be unique
        if not any(np.diff(dates) == dt.timedelta(days=0)):
            break
    yoes_l: List[int] = []
    yoes_ap = yoes_l.append
    moes_l: List[int] = []
    moes_ap = moes_l.append
    does_l: List[int] = []
    does_ap = does_l.append
    for d in dates:
        yoes_ap(d.year)
        moes_ap(d.month)
        does_ap(d.day)
    yoes = np.array(yoes_l, dtype=np.int16)
    moes = np.array(moes_l, dtype=np.int8)
    does = np.array(does_l, dtype=np.int8)
    del yoes_l
    del moes_l
    del does_l
    tes = np.array(amp, dtype=np.float32)
    return yoes, moes, does, tes


def from_json(table: dict, generator: create.Generate) -> create.Generate:
    """Create dates and total emissions from dictionary.

    Data is originally set in a json file, and read using the `json` library.

    Parameters
    ----------
    table: dict
        The dates and emissions as handles in the dictionary.
    generator: create.Generate
        A Generate object to which the table values are written.

    Returns
    -------
    yoes: np.ndarray
        Array of length 'size' with the year of a date
    moes: np.ndarray
        Array of length 'size' with the month of a date
    does: np.ndarray
        Array of length 'size' with the day of a date
    tes:
        Array of length 'size' with the Total_Emission as a 1D numpy array

    Raises
    ------
    KeyError
        If the sections in the loaded json file do not all have the same length.
    """
    _AVAILABLE_KEYS = {
        "dates",
        "emissions",
        "lat",
        "lon",
        "minimum injection height",
        "maximum injection height",
    }
    lens = map(len, table.values())
    if len(set(lens)) != 1:
        raise KeyError("All sections in json file must have the same length.")
    if not {"dates", "emissions"}.issubset(set(table)):
        raise KeyError("The keys 'dates' and 'emissions' must be present.")
    yoes_l: List[int] = []
    yoes_ap = yoes_l.append
    moes_l: List[int] = []
    moes_ap = moes_l.append
    does_l: List[int] = []
    does_ap = does_l.append
    tes_l: List[float] = []
    tes_ap = tes_l.append
    for dates in table.pop("dates"):
        y, m, d = dates.split("-")
        yoes_ap(int(y))
        moes_ap(int(m))
        does_ap(int(d))
    generator.yoes = np.array(yoes_l, dtype=np.int16)
    generator.moes = np.array(moes_l, dtype=np.int8)
    generator.does = np.array(does_l, dtype=np.int8)
    for emissions in table.pop("emissions"):
        tes_ap(float(emissions))
    generator.tes = np.array(tes_l, dtype=np.float32)
    if lats := table.pop("lat", False):
        lat_l: List[float] = []
        lat_ap = lat_l.append
        for lat in lats:
            lat_ap(lat)
        generator.lats = np.array(lat_l, dtype=np.float32)
    if lons := table.pop("lon", False):
        lon_l: List[float] = []
        lon_ap = lon_l.append
        for lon in lons:
            lon_ap(lon)
        generator.lons = np.array(lon_l, dtype=np.float32)
    if miihss := table.pop("minimum injection height", False):
        miihs_l: List[float] = []
        miihs_ap = miihs_l.append
        for miihs in miihss:
            miihs_ap(miihs)
        generator.miihs = np.array(miihs_l, dtype=np.float32)
    if mxihss := table.pop("maximum injection height", False):
        mxihs_l: List[float] = []
        mxihs_ap = mxihs_l.append
        for mxihs in mxihss:
            mxihs_ap(mxihs)
        generator.mxihs = np.array(mxihs_l, dtype=np.float32)
    _RED = "\033[0;31m"
    _BOLD = "\033[1m"
    _END = "\033[0m"
    if table:
        print(
            f"{_RED}{_BOLD}WARNING: some keys in the json table were not used: "
            + f"{table.keys()} Remember that the available keys are\n\t"
            + f"{_AVAILABLE_KEYS}.{_END}"
        )
    return generator



"""Compare values in two .nc files.

A quick way of comparing all relevant information found in the forcing data file used by
CESM2 (or similar) with the values stored in the file with synthetic data. The latest
created file with synthetic data is used.
"""

import pprint

import xarray as xr

import volcano_cooking.helper_scripts.functions as fnc


def compare_nc_files() -> None:
    original_file = fnc.find_original()
    synthetic_file = fnc.find_last_output("nc")

    forcing = xr.open_dataset(original_file, decode_times=False)
    my_forcing = xr.open_dataset(synthetic_file, decode_times=False)
    files = [forcing, my_forcing]

    # Just np.arange(241), one integer for each volcano
    print("Eruption_Number")
    pprint.pprint([f["Eruption_Number"] for f in files])
    print("")
    # print(forcing)
    # print(forcing.data_vars.keys())
    var_list = [
        "Eruption",
        "VEI",
        "Year_of_Emission",
        "Month_of_Emission",
        "Day_of_Emission",
        "Latitude",
        "Longitude",
        "Total_Emission",
        "Maximum_Injection_Height",
        "Minimum_Injection_Height",
    ]
    for v in var_list:
        print(v)
        pprint.pprint([f.variables[v].data for f in files])
        print("")


def main():
    compare_nc_files()


if __name__ == "__main__":
    main()



"""CLI for the view_generated_forcing module.

This implements a wrapper for a function to be able to use it as a stand-alone program. It
takes a filename as a mandatory input parameter.
"""

import sys

import click

import volcano_cooking.helper_scripts.view_generated_forcing as v


@click.command()
@click.argument("filename", type=click.Path(exists=True), required=False)
@click.option(
    "--dark/--no-dark",
    "-k/-K",
    default=False,
    show_default=True,
    type=bool,
    help="Plot with dark background",
)
@click.option(
    "--width",
    "-w",
    default=2,
    show_default=True,
    type=int,
    help="Width of the plot in number of columns",
)
@click.option(
    "--style",
    "-t",
    default="connected",
    show_default=True,
    type=str,
    help="Plot style used to draw the data",
)
@click.option(
    "--style-options",
    "style_options",
    default=False,
    show_default=True,
    is_flag=True,
    help="Show the available styles and exit",
)
@click.option(
    "--save/--no-save",
    "-s/-S",
    default=False,
    show_default=True,
    type=bool,
    help="Save the plot",
)
def main(
    filename: str, width: int, style: str, style_options: bool, dark: bool, save: bool
):
    """View a plot of `filename`."""
    styles = {"connected", "bars"}
    if style_options:
        print(f"Available styles: {styles}")
        return
    if style not in styles:
        sys.exit(f"Style '{style}' not available. Available styles: {styles}")
    v.view_forcing(in_file=filename, width=width, style=style, dark=dark, save=save)


if __name__ == "__main__":
    main()



"""Main file to run the program."""

import os
import ssl
import subprocess
import sys
from typing import List, Optional

import click
import wget

import volcano_cooking.configurations.shift_eruption_to_date as shift_eruption_to_date
import volcano_cooking.synthetic_volcanoes as sv
from volcano_cooking import __version__


@click.command()
@click.version_option(version=__version__)
@click.option(
    "-o",
    "--option",
    count=True,
    show_default=True,
    help="Set the data creation option. "
    + "The default is option 0, -o gives option 1.",
)
@click.option(
    "frc",
    "-f",
    type=int,
    default=0,
    show_default=True,
    help="Choose what forcing generator to use. "
    + "See volcano-cooking --lst for a list of all available generators.",
)
@click.option(
    "init_year",
    "-init",
    type=int,
    default=[1850, 1, 15],
    show_default=True,
    multiple=True,
    help="Set the first model year. If used with 'shift-eruption-to-date', "
    + "this becomes the year of the eruption and three instances can be used to specify "
    + "year, month, and day of the eruption. ",
)
@click.option(
    "size",
    "-s",
    type=int,
    default=3000,
    show_default=True,
    help="Set the number of volcanoes to generate.",
)
@click.option(
    "--lst/--no-lst",
    default=False,
    show_default=True,
    type=bool,
    help="Show a list of available forcing generators.",
)
@click.option(
    "--shift-eruption-to-date",
    "shift_eruption",
    is_flag=False,
    flag_value=True,
    default=False,
    show_default=True,
    type=str,
    help="Shift eruptions of provided file to 'init_year'. "
    + "If no file is provided, the last created file is used.",
)
@click.option(
    "--run-ncl",
    "run_ncl",
    is_flag=True,
    default=False,
    show_default=True,
    type=bool,
    help="Run the shell script that generate forcing from emission file.",
)
@click.option(
    "--package-last",
    "package_last",
    is_flag=True,
    default=False,
    show_default=True,
    type=bool,
    help="Move all last created files to a 'source-file' directory.",
)
@click.option(
    "--file",
    "file",
    type=click.Path(exists=True),
    default=None,
    help="Set volcanic eruption dates and strength from a json file.",
)
def main(
    frc: int,
    init_year: List[int],
    size: int,
    lst: bool,
    option: int,
    shift_eruption: str,
    run_ncl: bool,
    package_last: bool,
    file: Optional[str],
) -> None:
    # First handle list, run ncl and package commands which overrides all other.
    if lst:
        for cl in sv.__GENERATORS__:
            print(f"{cl}: {sv.__GENERATORS__[cl].__name__}")
        return
    if run_ncl:
        this_dir = os.path.dirname(os.path.abspath(__file__))
        shell_file = f"{this_dir}/create_cesm_frc.sh"
        subprocess.call(["sh", shell_file, this_dir])
        return
    if package_last:
        this_dir = os.path.dirname(os.path.abspath(__file__))
        shell_file = f"{this_dir}/package_last.sh"
        subprocess.call(["sh", shell_file])
        return

    _init_year = [1850, 1, 15]
    _init_year[: len(init_year)] = init_year
    if shift_eruption == "False":
        if option == 1:
            location = os.path.join(os.getcwd(), "data", "originals")
            file = "VolcanEESMv3.11_SO2_850-2016_Mscale_Zreduc_2deg_c191125.nc"
            if not os.path.isfile(os.path.join(location, file)):
                get_forcing_file(file)
            file = "fv_1.9x2.5_L30.nc"
            if not os.path.isfile(os.path.join(os.getcwd(), file)):
                get_forcing_file(
                    file,
                    url="https://svn-ccsm-inputdata.cgd.ucar.edu/trunk/inputdata"
                    + "/atm/cam/coords/fv_1.9x2.5_L30.nc",
                    not_forcing=True,
                )
        if file is None:
            sv.create_volcanoes(
                size=size, init_year=_init_year[0], version=frc, option=option
            )
        else:
            sv.create_volcanoes(file=file)
    elif shift_eruption == "True":
        shift_eruption_to_date.shift_eruption_to_date(tuple(_init_year), None)
    else:
        shift_eruption_to_date.shift_eruption_to_date(tuple(_init_year), shift_eruption)


def get_forcing_file(
    file: str, url: Optional[str] = None, not_forcing: bool = False
) -> None:
    """Download the original forcing file.

    It will be placed in `data/originals` inside the current working directory.

    Parameters
    ----------
    file: str
        The filename the file to download will be saved as.
    url: str, optional
        Provide the full url for where to download the file.
    not_forcing: bool
        If some other file is downloaded, use generic text without file size warning.
    """
    here = os.path.join(os.getcwd(), "data", "originals")
    os.makedirs(here, exist_ok=True)
    print(f"{file} was not found in {here}.")
    query_ = "Do you want to download the original forcing file? (2.2 GB)"
    query = f"Do you want to download '{file}'?" if not_forcing else query_
    if click.confirm(query):
        url_ = (
            "https://svn-ccsm-inputdata.cgd.ucar.edu/trunk/inputdata/atm/cam/chem"
            + "/stratvolc/VolcanEESMv3.11_SO2_850-2016_Mscale_Zreduc_2deg_c191125.nc"
        )
        url = url_ if url is None else url
        ssl._create_default_https_context = ssl._create_unverified_context
        wget.download(url, os.path.join(here, file))
        print("")
        if os.path.isfile(os.path.join(here, file)):
            print(f"Successfully downloaded {os.path.join(here, file)}.")
        else:
            sys.exit(
                "Failed to place the file in the correct directory. "
                + f"You may check if the file '{file}' was downloaded somewhere else, "
                + f"and if so place it in {here}."
            )
    else:
        sys.exit("Okay, I won't download it. Exiting gracefully.")


if __name__ == "__main__":
    main()



"""Functions used across scripts in the helper_scripts module."""

import os
import sys
from typing import Optional


def find_original(filename: Optional[str] = None) -> str:
    """Find the original .nc file with volcanic forcing.

    If nothing is specified, the function assumes the file name is
        `volcan-eesm_global_2015_so2-emissions-database_v1.0.nc`
    Optionally, the file name can be specified (it is still assumed it can be found in
    data/originals/).

    Parameters
    ----------
    filename: str (optional)
        Name of the file inside the originals directory you want to look at

    Returns
    -------
    str
        The path and name of the original forcing file
    """
    orig_dir = os.path.join("data", "originals")
    if not os.path.isdir(orig_dir):
        os.makedirs(orig_dir)
    if filename is None:
        original_file = os.path.join(
            "data",
            "originals",
            "volcan-eesm_global_2015_so2-emissions-database_v1.0.nc",
        )
    else:
        if ".nc" not in filename:
            filename += ".nc"
        original_file = os.path.join(orig_dir, filename)
    if not os.path.isfile(original_file):
        sys.exit(f"Can't find file {original_file}")

    return original_file


def find_last_output(ext: str) -> str:
    """Search in the output folder for a given file type and return last found file.

    Parameters
    ----------
    ext: str
            The file ending

    Returns
    -------
    str
        The path and name of the last saved .nc file

    Raises
    ------
    ValueError
        If the extension provided is not recognised
    """
    if ext[0] != ".":
        if "." in ext:
            raise ValueError(
                f"{ext} is not a valid extension. Use '.npz', 'npz', '.nc' or 'nc'."
            )
        else:
            ext = f".{ext}"
    synth_dir = os.path.join("data", "output")
    if not os.path.isdir(synth_dir):
        os.makedirs(synth_dir)
    synth_files = [
        f
        for f in os.listdir(synth_dir)
        if os.path.isfile(os.path.join(synth_dir, f)) and ext in f
    ]
    if not synth_files:
        sys.exit("No output files found.")
    synth_files.sort()
    return os.path.join(synth_dir, synth_files[-1])


def find_file(file: str) -> str:
    """Check if the file is found.

    Parameters
    ----------
    file: str
        Full path (absolute or relative to where the function is called) and file name of
        a custom file to be used.

    Returns
    -------
    str
        The same str that was sent in

    Raises
    ------
    ValueError
        If the file cannot be found in the path where the python interpreter is run a
        ValueError is raised.
    """
    if os.path.isfile(file):
        return file
    else:
        raise ValueError(f"Cannot find {file} =  in path {os.getcwd()}]")



"""Script to parse forcing files with the `stratvolc` data attribute."""

import os
import sys

import matplotlib.pyplot as plt
import xarray as xr


def sum_over_space():
    """Sum over latitude and longitude.

    Raises
    ------
    TypeError
        If the file extension is not .nc.
    FileNotFoundError
        If the file does not exist.
    """
    # Check file path
    data = sys.stdin.readlines()
    filename = data[0].strip("\n")
    if not isinstance(filename, str) or filename[-3:] != ".nc":
        raise TypeError(f"Are you sure {filename} is a valid .nc file?")
    if not os.path.exists(filename):
        raise FileNotFoundError(f"Cannot find file named {filename}.")
    # base = filename[:-3]
    ds = xr.open_dataset(filename, decode_times=False, chunks={"time": 100})
    only_time = ds.sum(dim=("lat", "lon", "altitude"))
    ds.close()
    if hasattr(ds, "date") and hasattr(ds, "datesec"):
        times = (only_time.date.values + only_time.datesec.values / 3600 / 24) / 10000
        plt.plot(times, only_time.stratvolc, "-o")
    elif hasattr(ds, "date"):
        times = only_time.date / 10000
        plt.plot(times, only_time.stratvolc, "-o")
    else:
        plt.plot(only_time.stratvolc, "-o")
    plt.show()


if __name__ == "__main__":
    sum_over_space()



"""Create volcanic forcing as a Poisson process with power law amplitudes."""

from abc import ABC, abstractmethod
from typing import Tuple

import numpy as np
import scipy.stats as scp_stats
import superposedpulses


class FrcGenerator(ABC):
    @abstractmethod
    def get_frc(self) -> Tuple[np.ndarray, np.ndarray]:
        """Return time and a realisation of the synthetic forcing signal."""


class StdFrc(FrcGenerator):
    def __init__(self, fs: float = 12, total_pulses: int = 300) -> None:
        if total_pulses < 1:
            raise ValueError(f"Can't create empty arrays, {total_pulses} = .")
        self.gamma = 0.1
        self.my_frc = superposedpulses.StandardForcingGenerator()
        self.my_frc.set_amplitude_distribution(lambda k: self.__lomax_amp(k, 1.8))
        self.times = np.arange(0, (total_pulses + 1) / self.gamma, 1 / fs)

    def __lomax_amp(self, k: int, c: float) -> np.ndarray:
        return scp_stats.lomax.rvs(c, size=k)

    def get_frc(self) -> Tuple[np.ndarray, np.ndarray]:
        frc = self.my_frc.get_forcing(self.times, self.gamma)
        ta, amp = frc.arrival_times, frc.amplitudes
        mask = np.argsort(ta)
        ta[:] = ta[mask]
        amp[:] = amp[mask]
        return ta, amp


class Frc(FrcGenerator):
    def __init__(self, fs: float = 1, size: int = 9999) -> None:
        # Lomax
        self.fs = fs
        self.fpp = superposedpulses.PointModel(
            gamma=0.1, total_duration=size, dt=1 / self.fs
        )
        my_forcing_gen = superposedpulses.StandardForcingGenerator()
        my_forcing_gen.set_amplitude_distribution(lambda k: self.__lomax_amp(k, 1.8))
        self.fpp.set_custom_forcing_generator(my_forcing_gen)

    def __lomax_amp(self, k: int, c: float) -> np.ndarray:
        return scp_stats.lomax.rvs(c, size=k)

    def get_fpp(self) -> Tuple[np.ndarray, np.ndarray]:
        t, f = self.fpp.make_realization()
        return t, f

    def get_frc(self) -> Tuple[np.ndarray, np.ndarray]:
        self.fpp.make_realization()
        last_frc = self.fpp.get_last_used_forcing()
        ta, amp = last_frc.arrival_times, last_frc.amplitudes
        mask = np.argsort(ta)
        ta[:] = ta[mask]
        amp[:] = amp[mask]
        return ta, amp



"""Create synthetic volcanic eruptions.

This script creates a .nc file consisting of volcanic eruptions and the
corresponding time of arrival down to the nearest day and the magnitude of the
eruption.

The lat/lon location is also included, but considered unimportant and thus the
same location is used for all volcanoes.
"""

from typing import Optional

import volcano_cooking.modules.create as create

# ====================================================================================== #
# Need:
#
# Dimension(s):
## Eruption number (do not contain variable attributes)
#
# Variable(s):
## Eruption (dim: Eruption_Number)
## -  Volcano number (not important, used to classify real volcanoes)
## -  Volcano name (not important)
## -  Notes and references (not important)
## VEI (dim: Eruption_Number)
## Year_of_Emission (dim: Eruption_Number)
## Month_of_Emission (dim: Eruption_Number)
## Day_of_Emission (dim: Eruption_Number)
## Latitude (dim: Eruption_Number)
## Longitude (dim: Eruption_Number)
## Total_Emission (dim: Eruption_Number)
## Maximum_Injection_Height (dim: Eruption_Number)
## Minimum_Injection_Height (dim: Eruption_Number)
# ====================================================================================== #


__GENERATORS__ = {
    0: create.GenerateRandomNormal,
    1: create.GenerateFPP,
    2: create.GenerateSingleVolcano,
    3: create.GenerateRegularIntervals,
    4: create.GenerateFromFile,
}


def create_volcanoes(
    size: int = 251,
    init_year: int = 1850,
    version: int = 0,
    option: int = 0,
    file: Optional[str] = None,
) -> None:
    """Create volcanoes starting at the year 1850.

    This will re-create the same data as can be found in forcing files used within the
    CESM2 general circulation model.

    Parameters
    ----------
    size: int
        The total number of eruptions
    init_year: int
        First year in the climate model
    version: int
        Choose one of the versions from the '__GENERATORS__' dictionary
    option: int
        Choose which option to use when generating forcing
    file: str, optional
        Read eruption dates and emissions from file.

    Raises
    ------
    IndexError
        If `version` is not a valid index of the generator dictionary.
    """
    # CREATE DATA -------------------------------------------------------------------- #

    if version not in __GENERATORS__ or version < 0:
        raise IndexError(
            f"No version exists for index {version}. "
            + "It must be one of {__GENERATORS__.keys()}."
        )
    if file is not None:
        print(f"Generating with '{__GENERATORS__[4].__name__}'...")
        print(
            "WARNING: When generating with this option, eruptions are not shifted to "
            + "have one eruption prior to the first simulation year. Add an extra "
            + "eruption at an early (and late) time that you know will cover the whole "
            + "simulation you are planning."
        )
        g = __GENERATORS__[4](size, init_year, file)
    else:
        print(f"Generating with '{__GENERATORS__[version].__name__}'...")
        g = __GENERATORS__[version](size, init_year)
    g.generate()
    all_arrs = g.get_arrays()

    # CREATE NETCDF FILE AND SAVE ---------------------------------------------------- #

    frc_cls = create.ReWrite(*all_arrs) if option == 1 else create.Data(*all_arrs)
    frc_cls.make_dataset()
    frc_cls.save_to_file()


def main():
    create_volcanoes(size=300, init_year=1)


if __name__ == "__main__":
    main()



"""Print out the data from all ten variables in a table, starting at any given year."""

import xarray as xr

import volcano_cooking.helper_scripts.functions as fnc


def inspect_year() -> None:
    year = 1865

    file = fnc.find_last_output("nc")
    print(file)

    f = xr.open_dataset(file, decode_times=False)

    var_list = [
        "Eruption",
        "VEI",
        "Year_of_Emission",
        "Month_of_Emission",
        "Day_of_Emission",
        "Latitude",
        "Longitude",
        "Total_Emission",
        "Maximum_Injection_Height",
        "Minimum_Injection_Height",
    ]
    var_list_short = [
        "Eru",
        "VEI",
        "YoE",
        "MoE",
        "DoE",
        "Lat",
        "Lon",
        "TE",
        "MaxIH",
        "MinIH",
    ]

    col_width = 12
    print("".join(element.ljust(col_width) for element in var_list_short))
    for i, y in enumerate(f.variables["Year_of_Emission"].data):
        if y <= year:
            p_list = [str(f.variables[var_list[j]].data[i])[:10] for j in range(10)]
            print("".join(element.ljust(col_width) for element in p_list))


def main():
    inspect_year()


if __name__ == "__main__":
    main()



"""Shift the second eruption to a given date.

Note
----
This script is intended to be used with the `--shift-eruption-to-date` flag. If no
'synthetic_volcanoes' file is present, it will create one and try and shift the eruption
generated in that file, before a second file is saved with the shifted eruption.
"""

from typing import Optional, Tuple

import numpy as np
import xarray as xr

import volcano_cooking.helper_scripts.functions as fnc
import volcano_cooking.modules.create as create


def shift_eruption_to_date(
    eruption_date: Tuple[int, int, int], file: Optional[str]
) -> None:
    """Shift the eruption date of the synthetic volcanoes file.

    Parameters
    ----------
    eruption_date: Tuple[int, int, int]
        Tuple with the new eruption date.
    file: str
        Full path (absolute or relative to where the function is called) and file name of
        a custom file to be used.

    Raises
    ------
    ValueError
        If the eruption date is not a valid date or if it is before or after the first and
        last eruptions.
    """
    if eruption_date[0] < 0 or eruption_date[0] > 9999:
        raise ValueError("Year of eruption must be between 0 and 9999, inclusive.")
    if eruption_date[1] < 1 or eruption_date[1] > 12:
        raise ValueError("Month of eruption must be between 1 and 12, inclusive.")
    if eruption_date[2] < 1 or eruption_date[2] > 28:
        raise ValueError("Day of eruption must be between 1 and 28, inclusive.")
    arrs = open_file("nc", file)
    first = arrs[1][0] * 10000 + arrs[2][0] * 100 + arrs[3][0]
    last = arrs[1][-1] * 10000 + arrs[2][-1] * 100 + arrs[3][-1]
    new = eruption_date[0] * 10000 + eruption_date[1] * 100 + eruption_date[2]
    if new < first or new > last:
        raise ValueError(
            f"The eruption date ({new}) is not in between "
            + f"the first and last eruption ({first}, {last})."
        )
    arrs[1][1] = eruption_date[0]
    arrs[2][1] = eruption_date[1]
    arrs[3][1] = eruption_date[2]
    frc_cls = create.Data(*arrs)
    frc_cls.make_dataset()
    frc_cls.save_to_file()


def open_file(ext: str, in_file: Optional[str] = None) -> Tuple[np.ndarray, ...]:
    """Load variables from the last saved file of given extension.

    Parameters
    ----------
    ext: str
        Extension of the file that should be used
    in_file: str, optional
        Full path (absolute or relative to where the function i called) and file name of a
        custom file to be used.

    Returns
    -------
    np.ndarray, ...
        All arrays saved in the synthetic_volcanoes file, same format as the input of the
        `create.Data` class.

    Raises
    ------
    NameError
        If the extension is not either 'npz' or 'nc', no files can be found and the
        variables 'yoes', 'moes', 'does' and 'tes' cannot be found.
    """
    if in_file is None:
        file = fnc.find_last_output(ext)
    elif in_file.split(".")[-1] == "npz":
        file = fnc.find_file("".join(in_file.split(".")[:-1]) + ".nc")
    else:
        file = fnc.find_file(in_file)
    if "nc" not in ext:
        raise NameError(
            "Data arrays cannot be found. "
            + f"Extension was {ext} = , but should be `nc`."
        )

    with xr.open_dataset(file, decode_times=False) as f:
        yoes = f.variables["Year_of_Emission"].data
        moes = f.variables["Month_of_Emission"].data
        does = f.variables["Day_of_Emission"].data
        tes = f.variables["Total_Emission"].data
        eruptions = f.variables["Eruption"].data
        lats = f.variables["Latitude"].data
        lons = f.variables["Longitude"].data
        veis = f.variables["VEI"].data
        miihs = f.variables["Minimum_Injection_Height"].data
        mxihs = f.variables["Maximum_Injection_Height"].data
    return eruptions, yoes, moes, does, lats, lons, tes, veis, miihs, mxihs



from volcano_cooking.modules.create.create_data import *  # noqa:F401,F403
from volcano_cooking.modules.create.create_dates import *  # noqa:F401,F403
from volcano_cooking.modules.create.create_frc import *  # noqa:F401,F403
from volcano_cooking.modules.create.rewrite_frc_file import *  # noqa:F401,F403

# See https://github.com/RaRe-Technologies/gensim/issues/1551
# Another option is:
# __all__ = ["random_dates",]



"""Add attributes to `altitude_int` coordinate.

Notes
-----
This script is run from the shell script `create_cesm_frc.sh`, but can be run manually as
```python
echo <input-file> | python src/volcano_cooking/modules/create/easy_fix.py
```
where `<input-file>` is the name of the netCDF file you want to edit.
"""

import os
import sys

import xarray as xr


def add_attributes() -> None:
    # Read from standard input
    data = sys.stdin.readlines()
    new_path = data[0].strip("\n")
    if not isinstance(new_path, str) or new_path[-3:] != ".nc":
        raise TypeError(f"Are you sure {new_path} is a valid netCDF file?")
    if not os.path.exists(new_path):
        raise FileNotFoundError(f"Cannot find file named {new_path}.")

    f_new = xr.open_dataset(new_path, decode_times=False)

    f_new["altitude_int"] = f_new["altitude_int"].assign_attrs(
        **{
            "standard_name": "altitude interval",
            "units": "km",
            "long_name": "altitude interval",
        }
    )
    encoding = {
        "lat": {"_FillValue": None},
        "lon": {"_FillValue": None},
        "altitude": {"_FillValue": None},
        "altitude_int": {"_FillValue": None},
        "stratvolc": {"_FillValue": 9.96921e36},
        "date": {"_FillValue": -2147483647},
        "datesec": {"_FillValue": -2147483647},
    }
    f_new.to_netcdf(
        f"{new_path[:-3]}_2.0.nc",
        "w",
        format="NETCDF3_64BIT",
        encoding=encoding,
    )


if __name__ == "__main__":
    add_attributes()



"""Script to change the volcanic forcing used in CESM.

Need to reset the time dimension, and the variables date, datesec and stratvolc.

`time` is easy, doesn't even contain any coordinates
- Set to `np.arange(N)`
`date` are ints and formatted in YYYYMMDD
- Choose using implementation from `volcano-cooking`
`datesec` are ints with seconds since midnight info
- Choose noon (43200), for example

`stratvolc` is a bit more tricky, but not too bad. Set all events to some lat/lon (0, 0),
and find a suitable altitude (perhaps from `volcano-cooking`). Then the values are set
using the implementations from `volcano-cooking`.
"""

import os
from datetime import datetime

import numpy as np
import xarray as xr

# from volcano_cooking import helper_scripts
from volcano_cooking.modules import convert
from volcano_cooking.modules.create.create_data import Data


class ReWrite(Data):
    def make_dataset(self) -> None:
        """Re-writes the original netCDF file with new variables.

        Raises
        ------
        IndexError
            If the loaded dataset does not include needed dimensions and variables.
        FileNotFoundError
            If the original forcing file cannot be found.
        """
        file = os.path.join(
            "data",
            "originals",
            "VolcanEESMv3.11_SO2_850-2016_Mscale_Zreduc_2deg_c191125",
        )
        if not os.path.exists(f"{file}.nc"):
            raise FileNotFoundError(
                f"{file} not found. Consult README on how to download."
            )
        f_orig = xr.open_dataset(f"{file}.nc", decode_times=False)
        # Check that the file contain crucial data.
        needed_dims = ["time", "altitude", "lat", "lon"]
        if any(d not in f_orig.dims for d in needed_dims):
            raise IndexError(f"Need dimensions {needed_dims}, found {f_orig.dims}")
        if "stratvolc" not in f_orig.data_vars:
            raise IndexError("Could not find 'stratvolc' variable.")
        if f_orig["stratvolc"].dims != ("time", "altitude", "lat", "lon"):
            raise IndexError(
                "'stratvolc' variable do not contain all needed dimensions, "
                + "or they are in incorrect order."
            )
        size = len(self.yoes)
        _, d_alt, d_lat, d_lon = f_orig["stratvolc"].shape
        new_eruptions = np.zeros((size, d_alt, d_lat, d_lon), dtype=np.float32)
        # We keep all spatial dimensions the same, but delete and re-sets the temporal
        # dimension.
        self.my_frc = xr.Dataset()
        self.miihs: np.ndarray
        self.mxihs: np.ndarray
        self.tes: np.ndarray
        # NOTE: which altitudes should we choose? Use bound from min and max injection
        # height, and the same emission on all altitudes. The emission at a given time
        # changes with altitude, but is generally of the same magnitude. Probably okay.
        zero_lat = np.abs(f_orig.lat.data).argmin()  # Find index closest to lat = 0
        zero_lon = np.abs(f_orig.lon.data).argmin()  # Find index closest to lon = 0
        self.miihs, self.mxihs = convert.adjust_altitude_range(
            self.miihs, self.mxihs, self.tes
        )
        self.tes = convert.adjust_emissions(
            self.miihs,
            self.mxihs,
            self.tes,
            np.array([zero_lon]),
            np.array([zero_lat]),
        )
        ai_dim = "altitude_int"
        self.my_frc = self.my_frc.expand_dims(**{ai_dim: f_orig[ai_dim]})
        self.my_frc = self.my_frc.assign_attrs(**f_orig.attrs)
        self.my_frc[ai_dim] = self.my_frc[ai_dim].assign_attrs(**f_orig[ai_dim].attrs)
        self.my_frc.attrs["data_summary"] = ""
        # Loops over time to set all spatial dimensions
        reset = False
        for i, emission in enumerate(self.tes):
            alt_range = f_orig.altitude.where(
                (f_orig.altitude <= self.mxihs[i]) & (f_orig.altitude >= self.miihs[i]),
                drop=True,
            ).astype(int)
            if not any(alt_range.data):
                if not reset:
                    txt_0 = "Re-setting alt. range for eruption at time:\n"
                    self.my_frc.attrs["data_summary"] += txt_0
                    reset = True
                alt_range = f_orig.altitude[-10:].astype(int)
                nd = (
                    10000 * self.yoes[i].astype(np.float32)
                    + 100 * self.moes[i].astype(np.float32)
                    + self.does[i].astype(np.float32)
                )
                d = "0" * (8 - len(str(int(nd)))) + str(int(nd))
                amin_0 = f"{self.miihs[i]:.3f}"
                amax_0 = f"{self.mxihs[i]:.3f}"
                amin_1 = f"{alt_range[0].data}"
                amax_1 = f"{alt_range[-1].data}"
                txt = f"{d}: ({amin_0}, {amax_0}) -> ({amin_1}, {amax_1})\n"
                self.my_frc.attrs["data_summary"] += txt
                self.mxihs[i] = alt_range[-1]
                self.miihs[i] = alt_range[0]
            new_eruptions[i, alt_range, zero_lat, zero_lon] = emission
        new_dates = (
            10000 * self.yoes.astype(np.float32)
            + 100 * self.moes.astype(np.float32)
            + self.does.astype(np.float32)
        )
        new_datesecs = np.array([43200.0 for _ in range(size)])  # Noon

        for v in f_orig.data_vars:
            # Place variables in the new Dataset that are taken from the original. This
            # way, all meta data is also used in the new Dataset as in the old, which is
            # needed.
            if v == "stratvolc":
                the_input = xr.DataArray(
                    new_eruptions,
                    dims=["time", "altitude", "lat", "lon"],
                    coords={
                        "altitude": f_orig.altitude,
                        "lat": f_orig.lat,
                        "lon": f_orig.lon,
                    },
                )
            elif v == "date":
                the_input = xr.DataArray(new_dates.astype(np.int32), dims="time")
            elif v == "datesec":
                the_input = xr.DataArray(new_datesecs.astype(np.int32), dims="time")
            else:
                raise IndexError(
                    f"'{v}' is an unknown variable that I do not have an implementation for."
                )
            the_input = the_input.assign_attrs(**f_orig[v].attrs)
            self.my_frc = self.my_frc.assign({v: the_input})

        # helper_scripts.compare_datasets(self.my_frc, f_orig)

    def __set_global_attrs(self, file) -> None:
        for a in self.my_frc.attrs:
            if a == "filename":
                self.my_frc.attrs[a] = file
            elif a == "title":
                start = f"{self.yoes[0]}.{self.moes[0]}.{self.does[0]}"
                end = f"{self.yoes[-1]}.{self.moes[-1]}.{self.does[-1]}"
                self.my_frc.attrs[a] = (
                    "SO2 emissions from synthetic stratospheric "
                    + f"volcanoes, {start}-{end}"
                )
            elif a == "data_creator":
                self.my_frc.attrs[a] = (
                    "Eirik Rolland Enger, University of Tromsø, eirik.r.enger@uit.no"
                )
            elif a == "data_doi":
                self.my_frc.attrs[a] = "No doi yet"
            elif a == "data_source_url":
                self.my_frc.attrs[a] = "https://github.com/engeir/volcano-cooking"
            elif a == "data_reference":
                self.my_frc.attrs[a] = "No reference yet"
            elif a == "data_source_files":
                self.my_frc.attrs[a] = "https://github.com/engeir/volcano-cooking"
            elif a == "creation_date":
                if datetime.now().strftime("%Z"):
                    this_day = datetime.now().strftime("%a %b %d %X %Z %Y")
                else:
                    this_day = datetime.now().strftime("%a %b %d %X %Y")
                self.my_frc.attrs[a] = this_day
            elif a == "cesm_contact":
                self.my_frc.attrs[a] = "None"
            elif a == "data_script":
                self.my_frc.attrs[a] = (
                    "Generated with the 'volcano-cooking' CLI with the re-write option."
                )
            elif a == "data_summary":
                nd = (
                    10000 * self.yoes.astype(np.float32)
                    + 100 * self.moes.astype(np.float32)
                    + self.does.astype(np.float32)
                )
                tot_width = 40
                w_1, w_23, w_4, w_5 = 8, 7, 3, 11
                summary = "\nThis file is for the following volcanoes:\n"
                summary += "=" * tot_width + "\n"
                summary += "YYYYMMDD AltMin  AltMax  VEI Em(cm-3s-1)\n"
                for d, amin, amax, v, e in zip(
                    nd, self.miihs, self.mxihs, self.veis, self.tes
                ):
                    d_ = "0" * (w_1 - len(str(int(d)))) + str(int(d)) + " "
                    amin_ = f"{amin:.3f}".rjust(w_23) + " "
                    amax_ = f"{amax:.3f}".rjust(w_23) + " "
                    v_ = f"{str(v).center(w_4)} "
                    e_ = f"{e:.5e}".rjust(w_5)
                    nl = d_ + amin_ + amax_ + v_ + e_ + "\n"
                    summary += nl
                self.my_frc.attrs[a] += summary
                self.my_frc.attrs[a] = self.my_frc.attrs[a]

    def save_to_file(self) -> None:
        """Save the re-written forcing file with the date at the end.

        Raises
        ------
        ValueError
            If a dataset have not yet been created.
        """
        if not hasattr(self, "my_frc"):
            raise ValueError("You must make the dataset with 'make_dataset' first.")
        file = "VolcanEESMv3.11_SO2_850-2016_Mscale_Zreduc_2deg_c191125_edit"
        out_file = self.check_dir("nc", name=file)
        self.__set_global_attrs(file=out_file)
        encoding = {
            "lat": {"_FillValue": None},
            "lon": {"_FillValue": None},
            "altitude": {"_FillValue": None},
            "altitude_int": {"_FillValue": None},
            "stratvolc": {"_FillValue": 9.96921e36},
            "date": {"_FillValue": -2147483647},
            "datesec": {"_FillValue": -2147483647},
        }
        self.my_frc.to_netcdf(out_file, "w", format="NETCDF3_64BIT", encoding=encoding)



from volcano_cooking.modules.convert.adjust_emissions_and_heights import *  # noqa:F401,F403
from volcano_cooking.modules.convert.convert_between_variables import *  # noqa:F401,F403

# See https://github.com/RaRe-Technologies/gensim/issues/1551
# Another option is: (think I like this more, but seems the issuers gravitate to noqa)
# __all__ = ["vei_to_totalemission", "vei_to_injectionheights", "totalemission_to_vei"]



"""Compare two xarray datasets' dimensions and variables."""

import os
from typing import List, Union

import xarray as xr


def compare_datasets(*args: Union[str, xr.Dataset]) -> None:
    """Compare any number of xarray datasets.

    This function checks the dimensions; their names and if it has variable attributes,
    the variables; their size and their attributes.

    Parameters
    ----------
    args: Union[str, xr.Dataset]
        Any number of paths to netCDF files or xarray dataset objects

    Raises
    ------
    TypeError
        If the input parameters are not of type `str` or `xr.Dataset`
    """
    xr_list: List[xr.Dataset] = []
    xr_app = xr_list.append
    for arr in args:
        if isinstance(arr, str):
            if not os.path.isfile(arr):
                raise TypeError(f"{arr} is not a file")
            xr_app(xr.open_dataset(arr, decode_times=False))
        elif isinstance(arr, xr.Dataset):
            xr_app(arr)
        else:
            raise TypeError(
                f"Object is type {type(arr)}. Must be 'str' or 'xr.Dataset'."
            )

    print("Checking dimensions...")
    for i, arr in enumerate(xr_list):
        print(f"\tArray {i}:\t{arr.dims}")

    print("\nChecking coordinates...")
    for i, arr in enumerate(xr_list):
        print(f"\tArray {i}:\t{list(arr.coords)}")
        # print(f"\tArray {i}:\t{arr.coords}")

    print("\nChecking variables...")
    for i, arr in enumerate(xr_list):
        print(
            f"\tArray {i}:\t{[f'{v}, {arr[v].dims}, {arr[v].shape}' for v in arr.data_vars]}"
        )
        # print(f"\tArray {i}:\t{arr.data_vars}")
        for v in arr.data_vars:
            print(f"\t\t{list(arr[v].attrs)}")

    print("\nChecking attributes...")
    for i, arr in enumerate(xr_list):
        print(f"\tArray {i}:\t{list(arr.attrs)}")
